import re
import os
import functools
from itertools import chain
from datetime import date, datetime, timedelta

import yaml
from flask import current_app, Markup

from veripress import cache
from veripress.model.models import Page, Post, Widget
from veripress.model.parsers import get_standard_format_name, get_parser
from veripress.helpers import to_list, to_datetime, Pair, traverse_directory


class Storage(object):
    def __init__(self):
        """Initialization."""
        self._closed = False

    def close(self):
        """
        Close the storage.
        Subclasses should override this to close any file descriptor or database connection if necessary.
        """
        self._closed = True

    @property
    def closed(self):
        """
        Read-only property.
        This state should be changed only in 'close' method.
        """
        return self._closed

    @cache.memoize(timeout=1 * 60)
    def fix_relative_url(self, publish_type, rel_url):
        """
        Fix post or page relative url to a standard, uniform format.

        :param publish_type: publish type ('post' or 'page')
        :param rel_url: relative url to fix
        :return: tuple(fixed relative url or file path if exists or None if cannot recognize, file exists or not)
        :raise ValueError: unknown publish type
        """
        if publish_type == 'post':
            return self.fix_post_relative_url(rel_url), False
        elif publish_type == 'page':
            return self.fix_page_relative_url(rel_url)
        else:
            raise ValueError('Publish type "{}" is not supported'.format(publish_type))

    @staticmethod
    @cache.memoize(timeout=2 * 60 * 60)  # actually, it will never change
    def fix_post_relative_url(rel_url):
        """
        Fix post relative url to a standard, uniform format.

        Possible input:
        - 2016/7/8/my-post
        - 2016/07/08/my-post.html
        - 2016/8/09/my-post/
        - 2016/8/09/my-post/index
        - 2016/8/09/my-post/index.htm
        - 2016/8/09/my-post/index.html

        :param rel_url: relative url to fix
        :return: fixed relative url, or None if cannot recognize
        """
        m = re.match(
            r'^(?P<year>\d{4})/(?P<month>\d{1,2})/(?P<day>\d{1,2})/(?P<post_name>[^/]+?)'
            r'(?:(?:\.html)|(?:/(?P<index>index(?:\.html?)?)?))?$',
            rel_url
        )
        if not m:
            return None

        year, month, day, post_name = m.groups()[:4]
        try:
            d = date(year=int(year), month=int(month), day=int(day))
            return '/'.join((d.strftime('%Y/%m/%d'), post_name, 'index.html' if m.group('index') else ''))
        except (TypeError, ValueError):
            # the date is invalid
            return None

    @staticmethod
    def fix_page_relative_url(rel_url):
        """
        Fix page relative url to a standard, uniform format.

        Possible input:
        - my-page
        - my-page/
        - my-page/index
        - my-page/index.htm
        - my-page/index.html
        - my-page/specific.file

        NOTE!
        Because custom page has a very strong connection with the storage type chosen,
        this method should be implemented in subclasses.

        :param rel_url: relative url to fix
        :return: tuple(fixed relative url or file path if exists or None if cannot recognize, file exists or not)
        """
        raise NotImplementedError

    def get_posts(self, include_draft=False, filter_functions=None):
        """Get all posts, returns an iterable of Post object."""
        raise NotImplementedError

    def get_post(self, rel_url, include_draft=False):
        """Get post for given relative url, returns a Post object."""
        raise NotImplementedError

    def get_tags(self):
        """Get all tags as a list of dict_item(tag_name, Pair(count_all, count_published))."""
        raise NotImplementedError

    def get_categories(self):
        """Get all categories as a list of dict_item(category_name, Pair(count_all, count_published))."""
        raise NotImplementedError

    def get_pages(self, include_draft=False):
        """Get all custom pages, returns an iterable of Page object."""
        raise NotImplementedError

    def get_page(self, rel_url, include_draft=False):
        """Get custom page for given relative url, returns a Page object."""
        raise NotImplementedError

    def get_widgets(self, position=None, include_draft=False):
        """Get all widgets, returns an iterable of Widget object."""
        raise NotImplementedError

    @staticmethod
    def _filter_result(result, filter_functions=None):
        """
        Filter result with given filter functions.

        :param result: an iterable object
        :param filter_functions: some filter functions
        :return: a filter object (filtered result)
        """
        if filter_functions is not None:
            for filter_func in filter_functions:
                result = filter(filter_func, result)
        return result

    def get_posts_with_limits(self, include_draft=False, **limits):
        """
        Get all posts and filter them as needed.

        :param include_draft: return draft posts or not
        :param limits: other limits to the attrs of the result, should be a dict with string or list values
        :return: an iterable of Post objects
        """
        filter_funcs = []

        for attr in ('title', 'layout', 'author', 'email', 'tags', 'categories'):
            if limits.get(attr):
                filter_set = set(to_list(limits.get(attr)))

                def get_filter_func(filter_set_, attr_):
                    return lambda p: filter_set_.intersection(to_list(getattr(p, attr_)))

                filter_funcs.append(get_filter_func(filter_set, attr))

        for attr in ('created', 'updated'):
            interval = limits.get(attr)
            if isinstance(interval, (list, tuple)) and len(interval) == 2 \
                    and isinstance(interval[0], date) and isinstance(interval[1], date):
                # [start date(time), end date(time)]
                start, end = interval
                start = to_datetime(start)
                if not isinstance(end, datetime):
                    # 'end' is a date, we should convert it to 00:00:00 of the next day,
                    # so that posts of that day will be included
                    end = datetime.strptime('%04d-%02d-%02d' % (end.year, end.month, end.day), '%Y-%m-%d')
                    end += timedelta(days=1)

                def get_filter_func(attr_, start_dt, end_dt):
                    return lambda p: start_dt <= getattr(p, attr_) < end_dt

                filter_funcs.append(get_filter_func(attr, start, end))

        return self.get_posts(include_draft=include_draft, filter_functions=filter_funcs)

    def search_for(self, query, include_draft=False):
        """
        Search for a query text.

        :param query: keyword to query
        :param include_draft: return draft posts/pages or not
        :return: an iterable object of posts and pages (if allowed).
        """
        query = query.lower()
        if not query:
            return []

        def contains_query_keyword(post_or_page):
            contains = query in post_or_page.title.lower() \
                       or query in Markup(
                get_parser(post_or_page.format).parse_whole(post_or_page.raw_content)
            ).striptags().lower()
            return contains

        return filter(contains_query_keyword, chain(self.get_posts(include_draft=include_draft),
                                                    self.get_pages(include_draft=include_draft)
                                                    if current_app.config['ALLOW_SEARCH_PAGES'] else []))


class FileStorage(Storage):
    @staticmethod
    @cache.memoize(timeout=1 * 60)
    def fix_page_relative_url(rel_url):
        """
        Fix page relative url to a standard, uniform format.

        Possible input:
        - my-page
        - my-page/
        - my-page/index
        - my-page/index.htm
        - my-page/index.html
        - my-page/specific.file

        :param rel_url: relative url to fix
        :return: tuple(fixed relative url or FILE PATH if exists or None if cannot recognize, file exists or not)
        """
        rel_url = rel_url.lstrip('/')  # trim all heading '/'
        endswith_slash = rel_url.endswith('/')
        rel_url = rel_url.rstrip('/') + ('/' if endswith_slash else '')  # preserve only one trailing '/'
        if not rel_url or rel_url == '/':
            return None, False

        file_path = os.path.join(current_app.instance_path, 'pages', rel_url.replace('/', os.path.sep))
        if rel_url.endswith('/'):
            index_html_file_path = os.path.join(file_path, 'index.html')
            if os.path.isfile(index_html_file_path):
                # index.html exists
                return index_html_file_path, True
            return rel_url, False
        elif os.path.isfile(file_path):
            ext = os.path.splitext(file_path)[1][1:]
            if get_standard_format_name(ext) is not None:
                # is source of custom page
                if current_app.config['PAGE_SOURCE_ACCESSIBLE']:
                    return file_path, True
            else:
                # is other direct files
                return file_path, True
        elif os.path.isdir(file_path):
            return rel_url + '/', False

        sp = rel_url.rsplit('/', 1)
        m = re.match(r'(.+)\.html?', sp[-1])
        if m:
            sp[-1] = m.group(1) + '.html'
        else:
            sp[-1] += '.html'
        return '/'.join(sp), False

    @staticmethod
    @cache.memoize(timeout=1 * 60)
    def search_file(search_root, search_filename, instance_relative_root=False):
        """
        Search for a filename in a specific search root dir.

        :param search_root: root dir to search
        :param search_filename: filename to search (no extension)
        :param instance_relative_root: search root is relative to instance path or not
        :return: tuple(full_file_path, extension without heading dot)
        """
        if instance_relative_root:
            search_root = os.path.join(current_app.instance_path, search_root)
        file_path = None
        file_ext = None
        for file in os.listdir(search_root):
            filename, ext = os.path.splitext(file)
            if filename == search_filename and ext and ext != '.':
                file_path = os.path.join(search_root, filename + ext)
                file_ext = ext[1:]  # remove heading '.' (dot)
                break
        return file_path, file_ext

    # noinspection PyUnresolvedReferences
    search_instance_file = staticmethod(functools.partial(search_file.__func__, instance_relative_root=True))

    @staticmethod
    @cache.memoize(timeout=1 * 60)
    def read_file(file_path):
        """
        Read yaml head and raw body content from a file.

        :param file_path: file path
        :return: tuple(meta, raw_content)
        """
        with open(file_path, 'r', encoding='utf-8') as f:
            whole = f.read().strip()

        if whole.startswith('---'):
            # may has yaml meta info, so we try to split it out
            sp = re.split(r'-{3,}', whole.lstrip('-'), maxsplit=1)
            if len(sp) == 2:
                # do have yaml meta info, so we read it
                return yaml.load(sp[0]), sp[1].lstrip()
        return {}, whole

    @cache.memoize(timeout=2 * 60)
    def get_posts(self, include_draft=False, filter_functions=None):
        """
        Get all posts from filesystem.

        :param include_draft: return draft posts or not
        :param filter_functions: filter functions to apply BEFORE result being sorted
        :return: an iterable of Post objects (the first is the latest post)
        """

        def posts_generator(path):
            """Loads valid posts one by one in the given path."""
            if os.path.isdir(path):
                for file in os.listdir(path):
                    filename, ext = os.path.splitext(file)
                    format_name = get_standard_format_name(ext[1:])
                    if format_name is not None and re.match(r'\d{4}-\d{2}-\d{2}-.+', filename):
                        # the format is supported and the filename is valid, so load this post
                        post = Post()
                        post.format = format_name
                        post.meta, post.raw_content = FileStorage.read_file(os.path.join(path, file))
                        post.rel_url = filename.replace('-', '/', 3) + '/'
                        post.unique_key = '/post/' + post.rel_url
                        yield post

        posts_path = os.path.join(current_app.instance_path, 'posts')
        result = filter(lambda p: include_draft or not p.is_draft, posts_generator(posts_path))
        result = self._filter_result(result, filter_functions)

        return sorted(result, key=lambda p: p.created, reverse=True)

    @cache.memoize(timeout=2 * 60)
    def get_post(self, rel_url, include_draft=False):
        """
        Get post for given relative url from filesystem.

        Possible input:
        - 2017/01/01/my-post/
        - 2017/01/01/my-post/index.html

        :param rel_url: relative url
        :param include_draft: return draft post or not
        :return: a Post object
        """
        raw_rel_url = str(rel_url)
        if rel_url.endswith('/index.html'):
            rel_url = rel_url.rsplit('/', 1)[0] + '/'  # remove the trailing 'index.html'
        post_filename = rel_url[:-1].replace('/', '-')

        post_file_path, post_file_ext = FileStorage.search_instance_file('posts', post_filename)
        if post_file_path is None or post_file_ext is None or get_standard_format_name(post_file_ext) is None:
            # no such post
            return None

        # construct the post object
        post = Post()
        post.rel_url = raw_rel_url
        post.unique_key = '/post/' + rel_url  # 'rel_url' contains no trailing 'index.html'
        post.format = get_standard_format_name(post_file_ext)
        post.meta, post.raw_content = FileStorage.read_file(post_file_path)
        return post if include_draft or not post.is_draft else None

    @cache.memoize(timeout=5 * 60)
    def get_tags(self):
        """
        Get all tags and post count of each tag.

        :return: dict_item(tag_name, Pair(count_all, count_published))
        """
        posts = self.get_posts(include_draft=True)
        result = {}
        for post in posts:
            for tag_name in set(post.tags):
                result[tag_name] = result.setdefault(tag_name, Pair(0, 0)) + Pair(1, 0 if post.is_draft else 1)
        return list(result.items())

    @cache.memoize(timeout=5 * 60)
    def get_categories(self):
        """
        Get all categories and post count of each category.

        :return dict_item(category_name, Pair(count_all, count_published))
        """
        posts = self.get_posts(include_draft=True)
        result = {}
        for post in posts:
            for category_name in set(post.categories):
                result[category_name] = result.setdefault(category_name, Pair(0, 0)) \
                                        + Pair(1, 0 if post.is_draft else 1)
        return list(result.items())

    @cache.memoize(timeout=2 * 60)
    def get_pages(self, include_draft=False):
        """
        Get all custom pages (supported formats, excluding other files like '.js', '.css', '.html').

        :param include_draft: return draft page or not
        :return: an iterable of Page objects
        """

        def pages_generator(pages_root_path):
            for file_path in traverse_directory(pages_root_path, yield_dir=False):
                rel_path = os.path.relpath(file_path, pages_root_path)
                rel_path, ext = os.path.splitext(rel_path)
                if not ext or ext == '.' or get_standard_format_name(ext[1:]) is None:
                    continue  # pragma: no cover, it seems that coverage cannot recognize this line

                if rel_path.endswith(os.path.sep + 'index'):
                    rel_path = rel_path[:-len('index')]
                else:
                    rel_path += '.html'
                page = self.get_page(rel_path.replace(os.path.sep, '/'), include_draft=include_draft)
                if page is not None:
                    yield page

        pages_path = os.path.join(current_app.instance_path, 'pages')
        return list(pages_generator(pages_path))

    @cache.memoize(timeout=2 * 60)
    def get_page(self, rel_url, include_draft=False):
        """
        Get custom page for given relative url from filesystem.

        Possible input:
        - my-page/
        - my-page/index.html
        - my-another-page.html
        - a/b/c/
        - a/b/c/d.html

        :param rel_url: relative url
        :param include_draft: return draft page or not
        :return: a Page object
        """
        page_dir = os.path.dirname(rel_url.replace('/', os.path.sep))
        page_path = os.path.join(current_app.instance_path, 'pages', page_dir)
        if not os.path.isdir(page_path):
            # no such directory
            return None

        page_filename = rel_url[len(page_dir):].lstrip('/')
        if not page_filename:
            page_filename = 'index'
        else:
            page_filename = os.path.splitext(page_filename)[0]

        page_file_path, page_file_ext = FileStorage.search_file(page_path, page_filename)
        if page_file_path is None or page_file_ext is None or get_standard_format_name(page_file_ext) is None:
            # no such page
            return None

        page = Page()
        page.rel_url = rel_url
        page.unique_key = '/' + (rel_url.rsplit('/', 1)[0] + '/' if rel_url.endswith('/index.html') else rel_url)
        page.format = get_standard_format_name(page_file_ext)
        page.meta, page.raw_content = FileStorage.read_file(page_file_path)
        return page if include_draft or not page.is_draft else None

    @cache.memoize(timeout=2 * 60)
    def get_widgets(self, position=None, include_draft=False):
        """
        Get widgets for given position from filesystem.

        :param position: position or position list
        :param include_draft: return draft widgets or not
        :return: an iterable of Widget objects
        """

        def widgets_generator(path):
            """Loads valid widgets one by one in the given path."""
            if os.path.isdir(path):
                for file in os.listdir(path):
                    _, ext = os.path.splitext(file)
                    format_name = get_standard_format_name(ext[1:])
                    if format_name is not None:
                        # the format is supported, so load it
                        widget = Widget()
                        widget.format = format_name
                        widget.meta, widget.raw_content = FileStorage.read_file(os.path.join(path, file))
                        yield widget

        widgets_path = os.path.join(current_app.instance_path, 'widgets')
        positions = to_list(position) if position is not None else position
        result = filter(lambda w: (w.position in positions if positions is not None else True)
                                  and (include_draft or not w.is_draft),
                        widgets_generator(widgets_path))
        return sorted(result, key=lambda w: (w.position, w.order))



import os
import re
import shutil

import click

from veripress_cli import cli


def get_themes_dir():
    from veripress import app
    return os.path.join(app.instance_path, 'themes')


@cli.group(name='theme', short_help='Manage themes.', help='This set of sub-commands help you to manage themes.')
def theme_cli():
    pass


@theme_cli.command('list', short_help='List all themes installed.')
def list_command():
    themes = list(filter(lambda x: os.path.isdir(os.path.join(get_themes_dir(), x)), os.listdir(get_themes_dir())))
    if not themes:
        click.echo('No theme is installed.')
        return
    click.echo('Themes installed:')
    click.echo('\n'.join(themes))


@theme_cli.command('install', short_help='Install a theme from GitHub.',
                   help='This command will install theme from GitHub for you. '
                        'If you want to install a theme from the official veripress/themes repo, '
                        'enter things like "default" to specify the theme. '
                        'If you want to install a third-party theme, use "user/repo" format.\n\ne.g.\n\n'
                        '  $ veripress theme install default\n\n'
                        '  $ veripress theme install someone/the-theme --branch the-theme-branch\n\n'
                        'If you want to customize the theme name installed, use "--name" parameter.')
@click.argument('theme', nargs=1)
@click.option('--branch', '-b', default='master',
              help='From which branch to clone the theme. '
                   'Only matters when a string like "user/repo" is passed in as the theme name.')
@click.option('--name', '-n', help='Specify the theme name. If this is not specified, a default one will be used.')
def install_command(theme, branch, name):
    if re.fullmatch('[_\-A-Z0-9a-z]+', theme):
        theme_name = name or theme
        theme_path = os.path.join(get_themes_dir(), theme_name)
        cmd = 'git clone --branch {} https://github.com/veripress/themes.git {}'.format(theme, theme_path)
    else:
        m = re.fullmatch('([_\-A-Z0-9a-z]+)/([_\-A-Z0-9a-z]+)', theme)
        if not m:
            raise click.BadArgumentUsage('The theme should be like "default" (branch of veripress/themes) '
                                         'or "someone/the-theme" (third-party theme on GitHub)')
        user = m.group(1)
        repo = m.group(2)
        theme_name = name or repo
        theme_path = os.path.join(get_themes_dir(), theme_name)
        cmd = 'git clone --branch {} https://github.com/{}/{}.git {}'.format(branch, user, repo, theme_path)
    exit_code = os.system(cmd)
    if exit_code == 0:
        click.echo('\n"{}" theme has been installed successfully.'.format(theme_name))
    else:
        click.echo('\nSomething went wrong. Do you forget to install git? '
                   'Or is there another theme with same name existing?')


@theme_cli.command('uninstall', short_help='Uninstall a theme.')
@click.argument('theme', nargs=1)
def uninstall_command(theme):
    theme_path = os.path.join(get_themes_dir(), theme)
    if os.path.isdir(theme_path):
        shutil.rmtree(theme_path)
        click.echo('"{}" theme has been uninstalled successfully.'.format(theme))
    else:
        click.echo('There is no such theme.')


@theme_cli.command('update', short_help='Update a theme.')
@click.argument('theme', nargs=1)
def uninstall_command(theme):
    theme_path = os.path.join(get_themes_dir(), theme)
    if os.path.isdir(theme_path):
        cur_dir = os.getcwd()
        os.chdir(theme_path)
        exit_code = os.system('git pull')
        os.chdir(cur_dir)
        if exit_code == 0:
            click.echo('\n"{}" theme has been updated successfully.'.format(theme))
        else:
            click.echo('\nSomething went wrong. Do you forget to install git? Or did you modify the theme by yourself?')
    else:
        click.echo('There is no such theme.')



import os
import re
from collections import Iterable
from datetime import date, datetime, timedelta, timezone

import pytz


def url_rule(blueprint_or_app, rules, endpoint=None, view_func=None, **options):
    """
    Add one or more url rules to the given Flask blueprint or app.

    :param blueprint_or_app: Flask blueprint or app
    :param rules: a single rule string or a list of rules
    :param endpoint: endpoint
    :param view_func: view function
    :param options: other options
    """
    for rule in to_list(rules):
        blueprint_or_app.add_url_rule(rule, endpoint=endpoint, view_func=view_func, **options)


def to_list(item_or_list):
    """
    Convert a single item, a tuple, a generator or anything else to a list.

    :param item_or_list: single item or iterable to convert
    :return: a list
    """
    if isinstance(item_or_list, list):
        return item_or_list
    elif isinstance(item_or_list, (str, bytes)):
        return [item_or_list]
    elif isinstance(item_or_list, Iterable):
        return list(item_or_list)
    else:
        return [item_or_list]


def to_datetime(date_or_datetime):
    """
    Convert a date object to a datetime object, or return as it is if it's not a date object.

    :param date_or_datetime: date or datetime object
    :return: a datetime object
    """
    if isinstance(date_or_datetime, date) and not isinstance(date_or_datetime, datetime):
        d = date_or_datetime
        return datetime.strptime('%04d-%02d-%02d' % (d.year, d.month, d.day), '%Y-%m-%d')
    return date_or_datetime


def timezone_from_str(tz_str):
    """
    Convert a timezone string to a timezone object.

    :param tz_str: string with format 'Asia/Shanghai' or 'UTC±[hh]:[mm]'
    :return: a timezone object (tzinfo)
    """
    m = re.match(r'UTC([+|-]\d{1,2}):(\d{2})', tz_str)
    if m:
        # in format 'UTC±[hh]:[mm]'
        delta_h = int(m.group(1))
        delta_m = int(m.group(2)) if delta_h >= 0 else -int(m.group(2))
        return timezone(timedelta(hours=delta_h, minutes=delta_m))

    # in format 'Asia/Shanghai'
    try:
        return pytz.timezone(tz_str)
    except pytz.exceptions.UnknownTimeZoneError:
        return None


class ConfigurationError(Exception):
    """Raise this when there's something wrong with the configuration."""
    pass


class Pair(object):
    """A class that just represent two value."""

    __dict__ = ['first', 'second']

    def __init__(self, first=None, second=None):
        self.first = first
        self.second = second

    def __repr__(self):
        return '<{} ({}, {})>'.format(self.__class__.__name__, repr(self.first), repr(self.second))

    def __eq__(self, other):
        if isinstance(other, Pair):
            return self.first == other.first and self.second == other.second
        return super().__eq__(other)

    def __bool__(self):
        return bool(self.first) or bool(self.second)

    __nonzero__ = __bool__  # for Python 2.x

    def __add__(self, other):
        a, b = other
        return Pair(self.first + a, self.second + b)

    def __sub__(self, other):
        a, b = other
        return Pair(self.first - a, self.second - b)

    def __getitem__(self, item):
        if isinstance(item, int):
            if item == 0:
                return self.first
            elif item == 1:
                return self.second
        raise IndexError

    def __len__(self):
        return 2


def validate_custom_page_path(path):
    """
    Check if a custom page path is valid or not, to prevent malicious requests.

    :param path: custom page path (url path)
    :return: valid or not
    """
    sp = path.split('/')
    if '.' in sp or '..' in sp:
        return False
    return True


def traverse_directory(dir_path, yield_dir=False):
    """
    Traverse through a directory recursively.

    :param dir_path: directory path
    :param yield_dir: yield subdirectory or not (only files will be yielded if left False)
    :return: a generator
    """
    if not os.path.isdir(dir_path):
        return

    for item in os.listdir(dir_path):
        new_path = os.path.join(dir_path, item)
        if os.path.isdir(new_path):
            if yield_dir:
                yield new_path + os.path.sep
            yield from traverse_directory(new_path, yield_dir)
        else:
            yield new_path


def parse_toc(html_content):
    """
    Parse TOC of HTML content if the SHOW_TOC config is true.

    :param html_content: raw HTML content
    :return: tuple(processed HTML, toc list, toc HTML unordered list)
    """
    from flask import current_app
    from veripress.model.toc import HtmlTocParser

    if current_app.config['SHOW_TOC']:
        toc_parser = HtmlTocParser()
        toc_parser.feed(html_content)
        toc_html = toc_parser.toc_html(depth=current_app.config['TOC_DEPTH'],
                                       lowest_level=current_app.config['TOC_LOWEST_LEVEL'])
        toc = toc_parser.toc(depth=current_app.config['TOC_DEPTH'],
                             lowest_level=current_app.config['TOC_LOWEST_LEVEL'])
        return toc_parser.html, toc, toc_html
    else:
        return html_content, None, None



import os
import shutil


def copy_folder_content(src, dst):
    """Copy all content in src directory to dst directory. The src and dst must be existing."""
    for file in os.listdir(src):
        file_path = os.path.join(src, file)
        dst_file_path = os.path.join(dst, file)
        if os.path.isdir(file_path):
            shutil.copytree(file_path, dst_file_path)
        else:
            shutil.copyfile(file_path, dst_file_path)


def remove_folder_content(path, ignore_hidden_file=False):
    """Remove all content in the given folder."""
    for file in os.listdir(path):
        if ignore_hidden_file and file.startswith('.'):
            continue

        file_path = os.path.join(path, file)
        if os.path.isdir(file_path):
            shutil.rmtree(file_path)
        else:
            os.remove(file_path)


def makedirs(path, mode=0o777, exist_ok=False):
    """A wrapper of os.makedirs()."""
    os.makedirs(path, mode, exist_ok)



import re
from itertools import islice
from datetime import date

from flask import request, send_file

from veripress import site, cache
from veripress.api import ApiException, Error
from veripress.model import storage
from veripress.model.models import Base
from veripress.model.parsers import get_parser
from veripress.helpers import validate_custom_page_path, parse_toc


@cache.memoize(timeout=5 * 60)
def site_info():
    return site


def posts(year: int = None, month: int = None, day: int = None, post_name: str = None):
    args = {k: [x.strip() for x in v.split(',')] for k, v in request.args.items()}

    for key in ('include_draft', 'start', 'count'):
        # pop out items that should not be passed into the 'get_posts' method as 'limits'
        args.pop(key, None)
    fields = args.pop('fields', None)  # fields that the API user needs, a list or None

    for key in ('created', 'updated'):
        if key in args:
            try:
                interval = args[key]
                # should be ['2017-02-13', '2017-03-13'] if it's valid
                for i in range(2):
                    y, m, d = re.match('(\d{4})-(\d{1,2})-(\d{1,2})', interval[i]).groups()
                    interval[i] = date(year=int(y), month=int(m), day=int(d))
            except (IndexError, AttributeError, ValueError):
                raise ApiException(message='The "{}" argument is invalid, '
                                           'and it should be like "2017-02-13,2017-03-13".'.format(key),
                                   error=Error.INVALID_ARGUMENTS)

    result_posts = storage.get_posts_with_limits(include_draft=False, **args)  # get the post list here

    return_single_item = False
    rel_url_prefix = ''
    if year is not None:
        rel_url_prefix += '%04d/' % year
    if month is not None:
        rel_url_prefix += '%02d/' % month
    if day is not None:
        rel_url_prefix += '%02d/' % day
    if post_name is not None:
        rel_url_prefix += '%s/' % post_name
        return_single_item = True  # if a full relative url is given, we return just ONE post, instead of a list
    result_posts = filter(lambda p: p.rel_url.startswith(rel_url_prefix), result_posts)

    start = request.args.get('start', '')
    start = int(start) if start.isdigit() else 0
    count = request.args.get('count', '')
    count = int(count) if count.isdigit() else -1

    result_posts_list = []
    for post in islice(result_posts, start, start + count if count >= 0 else None):
        parser = get_parser(post.format)
        post_d = post.to_dict()
        del post_d['raw_content']
        if return_single_item:
            # if a certain ONE post is needed, we parse all content instead of preview
            post_d['content'] = parser.parse_whole(post.raw_content)
            post_d['content'], post_d['toc'], post_d['toc_html'] = parse_toc(post_d['content'])
        else:
            # a list of posts is needed, we parse only previews
            post_d['preview'], post_d['has_more_content'] = parser.parse_preview(post.raw_content)
        if fields is not None:
            # select only needed fields to return
            assert isinstance(fields, list)
            full_post_d = post_d
            post_d = {}
            for key in fields:
                if key in full_post_d:
                    post_d[key] = full_post_d[key]
        result_posts_list.append(post_d)

    if result_posts_list and return_single_item:
        return result_posts_list[0]
    else:
        return result_posts_list if result_posts_list else None


def tags():
    return [{'name': item[0], 'published': item[1].second} for item in storage.get_tags()]


def categories():
    return [{'name': item[0], 'published': item[1].second} for item in storage.get_categories()]


def pages(page_path):
    if not validate_custom_page_path(page_path):
        raise ApiException(error=Error.NOT_ALLOWED, message='The visit of path "{}" is not allowed.'.format(page_path))

    rel_url, exists = storage.fix_relative_url('page', page_path)
    if exists:
        file_path = rel_url
        return send_file(file_path)
    elif rel_url is None:  # pragma: no cover, it seems impossible to make this happen, see code of 'fix_relative_url'
        raise ApiException(error=Error.BAD_PATH, message='The path "{}" cannot be recognized.'.format(page_path))
    else:
        page_d = cache.get('api-handler.' + rel_url)
        if page_d is not None:
            return page_d  # pragma: no cover, here just get the cached dict

        page = storage.get_page(rel_url, include_draft=False)
        if page is None:
            raise ApiException(error=Error.RESOURCE_NOT_EXISTS)
        page_d = page.to_dict()
        del page_d['raw_content']
        page_d['content'] = get_parser(page.format).parse_whole(page.raw_content)

        cache.set('api-handler.' + rel_url, page_d, timeout=2 * 60)
        return page_d


@cache.cached(timeout=2 * 60)
def widgets():
    result_widgets = storage.get_widgets(position=request.args.get('position'), include_draft=False)
    result = []
    for widget in result_widgets:
        widget_d = widget.to_dict()
        del widget_d['raw_content']
        widget_d['content'] = get_parser(widget.format).parse_whole(widget.raw_content)
        result.append(widget_d)
    return result if result else None


def search():
    query = request.args.get('q', '').strip().lower()
    if not query:
        raise ApiException(error=Error.INVALID_ARGUMENTS, message='The "q" argument is missed or invalid.')

    start = request.args.get('start', '')
    start = int(start) if start.isdigit() else 0
    count = request.args.get('count', '')
    count = int(count) if count.isdigit() else -1

    def remove_raw_content_field(p):
        del p['raw_content']
        return p

    result = list(islice(map(remove_raw_content_field, map(Base.to_dict, storage.search_for(query))),
                         start, start + count if count >= 0 else None))
    return result if result else None



import os
import math
import shutil

import click

from veripress_cli import cli
from veripress_cli.helpers import copy_folder_content, remove_folder_content, makedirs


def get_deploy_dir():
    from veripress import app
    return os.path.join(app.instance_path, '_deploy')


@cli.command('generate', short_help='Generate static pages.',
             help='This command will generate all HTML pages of the chosen instance.')
@click.option('--app-root', default='/',
              prompt='Please enter the application root (used as prefix of generated url)',
              help='The application root of your VeriPress instance. '
                   'For example, if you want to access the site through "http://example.com/blog/", '
                   'then "/blog/" should be passed in as the app root.')
def generate_command(app_root):
    deploy_dir = get_deploy_dir()
    if not os.path.isdir(deploy_dir):
        os.mkdir(deploy_dir)
    else:
        # remove all existing non-hidden files and dirs
        remove_folder_content(deploy_dir, ignore_hidden_file=True)

    from veripress import app

    app.config['APPLICATION_ROOT'] = app_root

    # mark as 'GENERATING_STATIC_PAGES' status, so that templates can react properly, e.g. remove search bar
    app.config['GENERATING_STATIC_PAGES'] = True

    do_generate()
    click.echo('\nOK! Now you can find all static pages in the "_deploy" folder.')


@cli.command('clean', short_help='Clean all generated static pages.',
             help='This command will clean all generated HTML pages of the chosen instance.')
def clean_command():
    deploy_dir = get_deploy_dir()
    if os.path.isdir(deploy_dir):
        remove_folder_content(get_deploy_dir(), ignore_hidden_file=True)
    click.echo('All generated static pages have been cleaned.')


def do_generate():
    from veripress import app
    from veripress.model import storage

    deploy_dir = get_deploy_dir()

    # copy global static folder
    dst_static_folder = os.path.join(deploy_dir, 'static')
    if os.path.isdir(app.static_folder):
        shutil.copytree(app.static_folder, dst_static_folder)

    # copy theme static files
    makedirs(dst_static_folder, mode=0o755, exist_ok=True)
    copy_folder_content(app.theme_static_folder, dst_static_folder)

    # collect all possible urls (except custom pages)
    all_urls = {'/', '/feed.xml', '/atom.xml', '/archive/'}
    with app.app_context():
        posts = list(storage.get_posts(include_draft=False))

        index_page_count = int(math.ceil(len(posts) / app.config['ENTRIES_PER_PAGE']))
        for i in range(2, index_page_count + 1):  # ignore '/page/1/', this will be generated separately later
            all_urls.add('/page/{}/'.format(i))

        for post in posts:
            all_urls.add(post.unique_key)
            all_urls.add('/archive/{}/'.format(post.created.strftime('%Y')))
            all_urls.add('/archive/{}/{}/'.format(post.created.strftime('%Y'), post.created.strftime('%m')))

        tags = storage.get_tags()
        for tag_item in tags:
            all_urls.add('/tag/{}/'.format(tag_item[0]))

        categories = storage.get_categories()
        for category_item in categories:
            all_urls.add('/category/{}/'.format(category_item[0]))

    with app.test_client() as client:
        # generate all possible urls
        for url in all_urls:
            resp = client.get(url)
            file_path = os.path.join(get_deploy_dir(), url.lstrip('/').replace('/', os.path.sep))
            if url.endswith('/'):
                file_path += 'index.html'

            makedirs(os.path.dirname(file_path), mode=0o755, exist_ok=True)
            with open(file_path, 'wb') as f:
                f.write(resp.data)

        # generate 404 page
        resp = client.get('/post/this-is-a-page-that-never-gonna-exist'
                          '-because-it-is-a-post-with-wrong-url-format/')
        with open(os.path.join(deploy_dir, '404.html'), 'wb') as f:
            f.write(resp.data)

    if app.config['STORAGE_TYPE'] == 'file':
        generate_pages_by_file()


def generate_pages_by_file():
    """Generates custom pages of 'file' storage type."""
    from veripress import app
    from veripress.model import storage
    from veripress.model.parsers import get_standard_format_name
    from veripress.helpers import traverse_directory

    deploy_dir = get_deploy_dir()

    def copy_file(src, dst):
        makedirs(os.path.dirname(dst), mode=0o755, exist_ok=True)
        shutil.copyfile(src, dst)

    with app.app_context(), app.test_client() as client:
        root_path = os.path.join(app.instance_path, 'pages')
        for path in traverse_directory(root_path):
            rel_path = os.path.relpath(path, root_path)  # e.g. 'a/b/c/index.md'
            filename, ext = os.path.splitext(rel_path)  # e.g. ('a/b/c/index', '.md')
            if get_standard_format_name(ext[1:]) is not None:
                # is source of custom page
                rel_url = filename.replace(os.path.sep, '/') + '.html'  # e.g. 'a/b/c/index.html'
                page = storage.get_page(rel_url, include_draft=False)
                if page is not None:
                    # it's not a draft, so generate the html page
                    makedirs(os.path.join(deploy_dir, os.path.dirname(rel_path)), mode=0o755, exist_ok=True)
                    with open(os.path.join(deploy_dir, filename + '.html'), 'wb') as f:
                        f.write(client.get('/' + rel_url).data)
                if app.config['PAGE_SOURCE_ACCESSIBLE']:
                    copy_file(path, os.path.join(deploy_dir, rel_path))
            else:
                # is other direct files
                copy_file(path, os.path.join(deploy_dir, rel_path))



import json
import os

from flask import Flask, send_from_directory, current_app, abort
from werkzeug.exceptions import NotFound


class CustomFlask(Flask):
    """
    Customize the original Flask class to fit the needs of this VeriPress project.
    """

    def send_static_file(self, filename):
        """
        Send static files from the static folder in the current selected theme prior to the global static folder.

        :param filename: static filename
        :return: response object
        """
        if self.config['MODE'] == 'api-only':
            abort(404)  # if 'api-only' mode is set, we should not send static files

        theme_static_folder = getattr(self, 'theme_static_folder', None)
        if theme_static_folder:
            try:
                return send_from_directory(theme_static_folder, filename)
            except NotFound:
                pass
        return super(CustomFlask, self).send_static_file(filename)


def create_app(config_filename, instance_path=None):
    """
    Factory function to create Flask application object.

    :param config_filename: absolute or relative (rel to the instance path) filename of the config file
    :param instance_path: the instance path to initialize or run a VeriPress app
    :return: a Flask app object
    """
    app_ = CustomFlask(__name__,
                       instance_path=instance_path or os.environ.get('VERIPRESS_INSTANCE_PATH') or os.getcwd(),
                       instance_relative_config=True)
    app_.config.update(dict(STORAGE_TYPE='file',
                            THEME='default',
                            CACHE_TYPE='simple',
                            MODE='view-only',
                            ENTRIES_PER_PAGE=5,
                            FEED_COUNT=10,
                            SHOW_TOC=True,
                            TOC_DEPTH=3,
                            TOC_LOWEST_LEVEL=3,
                            ALLOW_SEARCH_PAGES=True,
                            PAGE_SOURCE_ACCESSIBLE=False))
    app_.config.from_pyfile(config_filename, silent=True)

    theme_folder = os.path.join(app_.instance_path, 'themes', app_.config['THEME'])
    app_.template_folder = os.path.join(theme_folder, 'templates')  # use templates in the selected theme's folder
    app_.theme_static_folder = os.path.join(theme_folder, 'static')  # use static files in the selected theme's folder
    app_.static_folder = os.path.join(app_.instance_path, 'static')  # global static folder

    return app_


app = create_app('config.py')

site = {
    'title': 'Untitled',
    'subtitle': 'Yet another VeriPress blog.',
    'root_url': 'http://example.com',
    'timezone': 'UTC+00:00'
}
try:
    with app.open_instance_resource('site.json', mode='r') as site_file:
        # load site meta info to the site object
        site.update(json.load(site_file))
except FileNotFoundError:
    pass


@app.route('/_webhook', methods=['POST'], strict_slashes=False)
def webhook():
    """
    Run a custom python script when requested.
    User can pull git repositories, log something to a file, or do anything else in there.

    :return: always 204 NO CONTENT
    """
    try:
        with current_app.open_instance_resource('webhook.py', 'r') as script_file:
            exec(script_file.read())  # if there is the 'webhook.py' script, we execute it's content
    except FileNotFoundError:
        pass
    return '', 204


from flask_caching import Cache
cache = Cache(app, config=app.config)  # create the cache object with the app's config

if app.config['MODE'] in ('mixed', 'api-only'):
    import veripress.api
    app.register_blueprint(api.api_blueprint, url_prefix='/api')

if app.config['MODE'] in ('mixed', 'view-only'):
    import veripress.view
    app.register_blueprint(view.view_blueprint)

import veripress.model



import click

from veripress_cli import cli


@cli.command('serve', short_help='Serve/Run the application.',
             help='This command will start an HTTP server to serve the application.')
@click.option('--host', '-h', default='127.0.0.1', help='Host to serve the app.')
@click.option('--port', '-p', default=8080, help='Port to serve the app.')
def serve_command(host, port):
    click.echo('Starting HTTP server...')
    click.echo('HTTP server started. Running on http://{}:{}/'.format(host, port))

    from veripress import app
    try:
        from gevent.wsgi import WSGIServer
        server = WSGIServer((host, port), app)
        server.serve_forever()
    except ImportError:
        app.run(host=host, port=port)


@cli.command('preview', short_help='Preview the application.',
             help='This command will start an HTTP server to preview the application. '
                  'You should never use this command to run the app in production environment.')
@click.option('--host', '-h', default='127.0.0.1', help='Host to preview the app.')
@click.option('--port', '-p', default=8080, help='Port to preview the app.')
@click.option('--debug', is_flag=True, default=False, help='Preview in debug mode.')
def preview_command(host, port, debug):
    from veripress import app
    app.debug = debug
    app.config['TEMPLATES_AUTO_RELOAD'] = True
    app.config['CACHE_TYPE'] = 'null'
    app.run(host=host, port=port, debug=debug)



from datetime import datetime

from flask import current_app, g
from werkzeug.local import LocalProxy

import veripress.model.storages
from veripress.model.models import Base
from veripress.helpers import ConfigurationError


def get_storage():
    """
    Get storage object of current app context, will create a new one if not exists.

    :return: a storage object
    :raise: ConfigurationError: storage type in current_app.config is not supported
    """
    storage_ = getattr(g, '_storage', None)
    if storage_ is None:
        storage_type = current_app.config['STORAGE_TYPE']
        if storage_type == 'file':
            storage_ = g._storage = storages.FileStorage()
        else:
            raise ConfigurationError('Storage type "{}" is not supported.'.format(storage_type))
    return storage_


storage = LocalProxy(get_storage)

from veripress import app


@app.teardown_appcontext
def teardown_storage(e):
    """
    Automatically called when Flask tears down the app context.
    This will close the storage object created at the beginning of the current app context.
    """
    storage_ = getattr(g, '_storage', None)
    if storage_ is not None:
        storage_.close()


class CustomJSONEncoder(app.json_encoder):
    """
    Converts model objects to dicts, datetime to timestamp,
    so that they can be serialized correctly.
    """

    def default(self, obj):
        if isinstance(obj, Base):
            return obj.to_dict()
        elif isinstance(obj, datetime):
            return obj.strftime('%Y-%m-%d %H:%M:%S')
        return super(CustomJSONEncoder, self).default(obj)


app.json_encoder = CustomJSONEncoder  # use the customized JSON encoder when jsonify is called



from datetime import datetime

from veripress import site
from veripress.helpers import to_list, to_datetime


class Base(object):
    """Base model class, contains basic/general information of a post/page/widget."""

    def __init__(self):
        self.meta = {}
        self.raw_content = None
        self._format = None

    @property
    def format(self):
        return self._format

    @format.setter
    def format(self, value):
        if value is not None:
            self._format = value.lower()

    @property
    def is_draft(self):
        return self.meta.get('is_draft', False)

    def to_dict(self):
        """Convert attributes and properties to a dict (so that it can be serialized)."""
        return {k: getattr(self, k) for k in filter(lambda k: not k.startswith('_') and k != 'to_dict', dir(self))}

    def __eq__(self, other):
        if isinstance(other, Base):
            return self.to_dict() == other.to_dict()
        return super(Base, self).__eq__(other)


class AuthorMixIn(object):
    """Mix in author's name and email."""

    @property
    def author(self):
        return getattr(self, 'meta', {}).get('author', site.get('author'))

    @property
    def email(self):
        return getattr(self, 'meta', {}).get('email', site.get('email'))


class DateMixIn(object):
    """Mix in created data and updated date."""

    @property
    def created(self):
        return to_datetime(getattr(self, 'meta', {}).get('created'))

    @property
    def updated(self):
        return to_datetime(getattr(self, 'meta', {}).get('updated', self.created))


class TagCategoryMixIn(object):
    """Mix in tags and categories."""

    @property
    def tags(self):
        return to_list(getattr(self, 'meta', {}).get('tags', []))

    @property
    def categories(self):
        return to_list(getattr(self, 'meta', {}).get('categories', []))


class Page(Base, AuthorMixIn, DateMixIn):
    """Model class of publish type 'custom page' or 'page', with default layout 'page'."""
    _default_layout = 'page'

    def __init__(self):
        super().__init__()
        self.unique_key = None
        self.rel_url = None

    @property
    def layout(self):
        return self.meta.get('layout', self._default_layout)

    @property
    def title(self):
        result = self.meta.get('title')
        if result is None and self.rel_url:
            sp = self.rel_url.split('/')
            pos = len(sp) - 1
            while pos > 0 and (sp[pos] == 'index.html' or not sp[pos]):
                pos -= 1

            path_seg = sp[pos][:-len('.html')] if sp[pos].endswith('.html') else sp[pos]
            result = ' '.join(word[0].upper() + word[1:] for word in filter(lambda x: x, path_seg.split('-')))
        return result


class Post(Page, TagCategoryMixIn):
    """Model class of publish type 'post', with default layout 'post'."""
    _default_layout = 'post'

    @property
    def created(self):
        result = super(Post, self).created
        if result is None:
            d, _, _ = self.rel_url.rsplit('/', 2)
            result = datetime.strptime(d, '%Y/%m/%d')
        return result

    @property
    def title(self):
        result = self.meta.get('title')
        if result is None:
            _, post_name, _ = self.rel_url.rsplit('/', 2)
            result = ' '.join(word[0].upper() + word[1:] for word in filter(lambda x: x, post_name.split('-')))
        return result


class Widget(Base):
    """Model class of publish type 'widget'."""

    @property
    def position(self):
        return self.meta.get('position')

    @property
    def order(self):
        return self.meta.get('order')



import os
import re
from datetime import datetime

import click

from veripress_cli import cli
from veripress_cli.helpers import makedirs


def validate_repo_name(ctx, param, value):
    m = re.fullmatch('([._\-A-Z0-9a-z]+)/([._\-A-Z0-9a-z]+)', value)
    if not m:
        raise click.BadParameter('The repo name is invalid.')
    return m.groups()


@cli.command('setup-github-pages', short_help='Setup GitHub Pages.',
             help='This command will setup the deploy folder as a GitHub Pages repo.')
@click.option('--repo', '-r',
              prompt='Please enter your GitHub repo name, e.g. "someone/the-repo"',
              help='The GitHub repo name you want to deploy to.',
              callback=validate_repo_name)
@click.option('--name', '-n', prompt='Please enter your name (for git config)',
              help='Your name set in git config.')
@click.option('--email', '-e',
              prompt='Please enter your email (for git config)',
              help='Your email set in git config')
def setup_command(repo, name, email):
    user, repo = repo

    from veripress_cli.generate import get_deploy_dir
    deploy_dir = get_deploy_dir()
    makedirs(deploy_dir, mode=0o755, exist_ok=True)

    os.system('git -C "{}" init'.format(deploy_dir))
    os.system('git -C "{}" config user.email "{}"'.format(deploy_dir, email))
    os.system('git -C "{}" config user.name "{}"'.format(deploy_dir, name))
    os.system('git -C "{}" remote add origin git@github.com:{}/{}.git'.format(
        deploy_dir, user, repo))


@cli.command('deploy', short_help='Deploy GitHub Pages.',
             help='This command will deploy the static pages to GitHub Pages.')
def deploy_command():
    from veripress_cli.generate import get_deploy_dir
    deploy_dir = get_deploy_dir()
    makedirs(deploy_dir, mode=0o755, exist_ok=True)

    os.system('git -C "{}" add .'.format(deploy_dir))
    if os.system(
            'git -C "{}" diff --quiet --exit-code'.format(deploy_dir)) == 0 \
            and os.system(
        'git -C "{}" diff --quiet --cached --exit-code'.format(
            deploy_dir)) == 0:
        click.echo('There are no changes to be deployed.')
        return

    dt = datetime.now()
    os.system('git -C "{}" commit -m "Updated on {} at {}"'.format(
        deploy_dir, dt.strftime('%Y-%m-%d'), dt.strftime('%H:%M:%S')))
    os.system(
        'git -C "{}" push --set-upstream origin master'.format(deploy_dir))



import os
import functools
from itertools import chain

from flask import Blueprint, request, render_template, g

from veripress import site, cache
from veripress.model import storage
from veripress.model.parsers import get_parser
from veripress.helpers import url_rule, to_list

view_blueprint = Blueprint('view', __name__)


@view_blueprint.context_processor
def inject_context():
    """
    Inject some common objects into the context of templates.
    """
    return dict(site=site, storage=storage)


@view_blueprint.app_template_filter('content')
def parse_content_of_models(obj):
    """
    Parse the whole 'raw_content' attribute of a Post or Page or Widget object (in template files).

    :param obj: a Post or Page or Widget object
    :return: parsed whole content
    """
    return get_parser(obj.format).parse_whole(obj.raw_content)


@cache.memoize(timeout=2 * 60)
def custom_render_template(template_name_or_list, **context):
    """
    Try to render templates in the custom folder first, if no custom templates, try the theme's default ones.
    """
    response_str = render_template(
        functools.reduce(lambda x, y: x + [os.path.join('custom', y), y], to_list(template_name_or_list), []),
        **context
    )
    if hasattr(g, 'status_code'):
        status_code = g.status_code
    else:
        status_code = 200
    return response_str, status_code


def templated(template=None, *templates):
    """
    Decorate a view function with one or more default template name.
    This will try templates in the custom folder first, the theme's original ones second.

    :param template: template name or template name list
    """

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            template_ = template
            if template_ is None:
                template_ = request.endpoint.split('.', 1)[1].replace('.', '/') + '.html'
            context = func(*args, **kwargs)
            if context is None:
                context = {}
            elif not isinstance(context, dict):
                return context
            return custom_render_template(list(chain(to_list(template_), templates)), **context)

        return wrapper

    return decorator


@view_blueprint.errorhandler(404)
@templated('404.html')
def page_not_found(e):
    g.status_code = 404


from veripress.view import views

rule = functools.partial(url_rule, view_blueprint, methods=['GET'])

rule(['/feed.xml', '/atom.xml'], view_func=views.feed, strict_slashes=True)

rule(['/', '/page/<int:page_num>/'], view_func=views.index, strict_slashes=True)
rule('/post/<int:year>/<int:month>/<int:day>/<string:post_name>', view_func=views.post, strict_slashes=False)
rule('/category/<string:category_name>/', view_func=views.category, strict_slashes=True)
rule('/tag/<string:tag_name>/', view_func=views.tag, strict_slashes=True)
rule(['/archive/',
      '/archive/<int:year>/',
      '/archive/<int:year>/<int:month>/'], view_func=views.archive, strict_slashes=True)
rule('/search', view_func=views.search, strict_slashes=False)
rule('/<path:rel_url>', view_func=views.page, strict_slashes=True)



import re
import sys

try:
    from html.parser import HTMLParser
except ImportError:
    # Python 2.x
    # noinspection PyUnresolvedReferences,PyCompatibility
    from HTMLParser import HTMLParser


class _HtmlHeaderNode(object):
    """Represents a header element when parsing the HTML string."""

    def __init__(self, level):
        """Initialize attributes."""
        self.level = level  # header level of the element, e.g. 1 for <h1>, 2 for <h2>, etc
        self.id = ''  # anchor id (in-page link), used in 'id' and 'href' attribute of 'a' tag
        self.text = ''  # pure text content of header tag, e.g. 'Title' for '<h1>Title</h1>'
        self.inner_html = ''  # inner HTML
        self.father = None  # point to the direct father node
        self.children = []  # elements with lower levels that directly follows the current elem

    def to_dict(self):
        """Convert self to a dict object for serialization."""
        return {
            'level': self.level,
            'id': self.id,
            'text': self.text,
            'inner_html': self.inner_html,
            'children': [child.to_dict() for child in self.children]
        }


class HtmlTocParser(HTMLParser):
    """Parse table of content from a given HTML string."""

    # regular expression for replacing all punctuations in the header with '-'

    if sys.version.startswith('3.'):
        # Python 3.x, safe
        _punctuations_exp = re.compile(
            r'[\s\u0020-\u002f\u003a-\u0040\u005b-\u0060\u007b-\u007e\u00a0-\u00bf'
            r'\u2000-\u206f\u2e00-\u2e7f\u3000-\u303f\uff01-\uff0f\uff1a-\uff20'
            r'\uff3b-\uff40\uff5b-\uff65\uffe0-\uffe6\uffe8-\uffec\ufe10-\ufe1f]+'
        )
    else:
        # Python 2.x, DANGEROUS!!
        # don't know why but it does work WHEN and ONLY WHEN 'ur' leading is used...
        # and because 'ur' leading is not allowed in Python 3.x, so we use 'exec()' here
        # otherwise SyntaxError will be raised
        exec("""_punctuations_exp = re.compile(
            ur'[\s\u0020-\u002f\u003a-\u0040\u005b-\u0060\u007b-\u007e\u00a0-\u00bf'
            ur'\u2000-\u206f\u2e00-\u2e7f\u3000-\u303f\uff01-\uff0f\uff1a-\uff20'
            ur'\uff3b-\uff40\uff5b-\uff65\uffe0-\uffe6\uffe8-\uffec\ufe10-\ufe1f]+'
        )""")

    def __init__(self):
        """Initialize attributes."""
        if sys.version.startswith('3.'):
            # Python 3.x
            super().__init__(convert_charrefs=False)
        else:
            # use HTMLParser.__init__ because HTMLParser is an 'old' style class, which cannot be passed to super()
            # see http://codependentcodr.blogspot.com/2012/02/python-htmlparser-and-super.html
            HTMLParser.__init__(self)

        self._root = _HtmlHeaderNode(level=0)  # root node with no data of itself, only 'children' matters
        self._curr_node = self._root  # most recently handled header node
        self._in_header = False
        self._header_id_count = {}  # record header ids to avoid collisions
        self._html = ''  # full HTML string parsed
        self._temp_start_tag = ''  # temporary HTML start tag of this current header node

    def toc(self, depth=6, lowest_level=6):
        """
        Get table of content of currently fed HTML string.

        :param depth: the depth of TOC
        :param lowest_level: the allowed lowest level of header tag
        :return: a list representing the TOC
        """
        depth = min(max(depth, 0), 6)  # make depth >= 0 and <= 6
        depth = 6 if depth == 0 else depth  # depth == 0 <=> depth == 6
        lowest_level = min(max(lowest_level, 1), 6)  # make lowest_level >= 1 and <= 6
        toc = self._root.to_dict()['children']

        def traverse(curr_toc, dep, lowest_lvl, curr_depth=1):
            if curr_depth > dep:
                # clear all items of this depth and exit the recursion
                # curr_toc.clear()
                curr_toc[:] = []  # it's the same as the above line, for compatibility with Python 2.x
                return

            items_to_remove = []
            for item in curr_toc:
                if item['level'] > lowest_lvl:
                    # record item with low header level, for removing it later
                    items_to_remove.append(item)
                else:
                    traverse(item['children'], dep, lowest_lvl, curr_depth + 1)
            [curr_toc.remove(item) for item in items_to_remove]

        traverse(toc, depth, lowest_level)
        return toc

    def toc_html(self, depth=6, lowest_level=6):
        """
        Get TOC of currently fed HTML string in form of HTML string.

        :param depth: the depth of TOC
        :param lowest_level: the allowed lowest level of header tag
        :return: an HTML string
        """
        toc = self.toc(depth=depth, lowest_level=lowest_level)
        if not toc:
            return ''

        def map_toc_list(toc_list):
            result = ''
            if toc_list:
                result += '<ul>\n'
                result += ''.join(map(lambda x: u'<li>'
                                                u'<a href="#{}">{}</a>{}'
                                                u'</li>\n'.format(x['id'], x['inner_html'],
                                                                  map_toc_list(x['children'])),
                                      toc_list))
                result += '</ul>'
            return result

        return map_toc_list(toc)

    @property
    def html(self):
        """
        The parsed HTML string with additional a tag marking the header anchors.

        Example:
            `<h1>Title</h1>`

            becomes:

            `<h1><a id="Title" class="anchor"></a>Title</h1>`
        """
        return self._html

    @staticmethod
    def _get_level(tag):
        """Match the header level in the given tag name, or None if it's not a header tag."""
        m = re.match(r'^h([123456])$', tag, flags=re.IGNORECASE)
        if not m:
            return None
        return int(m.group(1))

    def handle_starttag(self, tag, attrs):
        curr_tag = u'<{}{}{}>'.format(tag, ' ' if attrs else '',
                                      ' '.join([u'{}="{}"'.format(*attr) for attr in attrs]))

        if self._in_header:
            self._curr_node.inner_html += curr_tag
            return

        level = self._get_level(tag)
        if level is not None:
            self._in_header = True
            new_node = _HtmlHeaderNode(level=level)
            while new_node.level <= self._curr_node.level:
                # the new node is a higher level header, e.g. new: h2, curr: h4
                # so we return back until the current is higher
                self._curr_node = self._curr_node.father
            new_node.father = self._curr_node  # assign the new node as a child of the current node
            self._curr_node.children.append(new_node)
            self._curr_node = new_node
            self._temp_start_tag = curr_tag
            return

        self._html += curr_tag

    def handle_startendtag(self, tag, attrs):
        curr_tag = u'<{}{}{} />'.format(tag, ' ' if attrs else '',
                                        ' '.join([u'{}="{}"'.format(*attr) for attr in attrs]))

        if self._in_header:
            self._curr_node.inner_html += curr_tag
            return

    def handle_endtag(self, tag):
        curr_tag = u'</{}>'.format(tag)

        if self._get_level(tag) is not None:
            header_id = self._punctuations_exp.sub('-', self._curr_node.text).strip('-')
            count = self._header_id_count.setdefault(header_id, 0)
            self._header_id_count[header_id] += 1
            if count > 0:
                header_id += '_%d' % count

            self._curr_node.id = header_id
            self._html += self._temp_start_tag  # start tag of the current header node
            self._html += u'<a id="{0}" href="#{0}" class="anchor"></a>'.format(header_id)  # anchor
            self._html += self._curr_node.inner_html + curr_tag  # header content and end tag
            self._temp_start_tag = ''
            self._in_header = False
            return
        elif self._in_header:
            self._curr_node.inner_html += curr_tag
            return

        self._html += curr_tag

    def handle_data(self, data):
        if self._in_header:
            self._curr_node.text += data
            self._curr_node.inner_html += data
            return

        self._html += data

    def handle_comment(self, data):
        # self._html += u'<!--{}-->'.format(data)
        self.handle_data(u'<!--{}-->'.format(data))

    def handle_entityref(self, name):
        # self._html += u'&{};'.format(name)
        self.handle_data(u'&{};'.format(name))

    def handle_charref(self, name):
        # self._html += u'&#{};'.format(name)
        self.handle_data(u'&#{};'.format(name))

    def error(self, message):
        pass  # pragma: no cover



import functools
from collections import Iterable, namedtuple

from flask import Blueprint, jsonify, Response, abort

from veripress.helpers import url_rule

api_blueprint = Blueprint('api', __name__)


class Error(object):
    """
    Defines API error codes and error messages.
    """

    _Error = namedtuple('Error', ('code', 'msg', 'status_code'))

    UNDEFINED = _Error(100, 'Undefined error.', 400)
    NO_SUCH_API = _Error(101, 'No such API.', 404)
    RESOURCE_NOT_EXISTS = _Error(102, 'The resource does not exist.', 404)
    INVALID_ARGUMENTS = _Error(103, 'Invalid argument(s).', 400)
    NOT_ALLOWED = _Error(104, 'The resource path is not allowed.', 403)
    BAD_PATH = _Error(105, 'The resource path cannot be recognized.', 400)


class ApiException(Exception):
    """Raised by API functions when something goes wrong."""

    def __init__(self, message=None, error=Error.UNDEFINED, status_code=None, payload=None):
        super(ApiException, self).__init__()
        self.message = message
        self.status_code = status_code
        self.error = error
        self.payload = payload

    def to_dict(self):
        result = dict(self.payload or {})
        result['code'] = self.error.code
        result['message'] = self.message or self.error.msg
        return result


@api_blueprint.errorhandler(ApiException)
def handle_api_exception(e):
    response = jsonify(e.to_dict())
    response.status_code = e.status_code or e.error.status_code
    return response


@api_blueprint.errorhandler(404)
def handle_page_not_found(e):
    return handle_api_exception(ApiException(error=Error.NO_SUCH_API))


def json_api(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        if result is None:
            raise ApiException(error=Error.RESOURCE_NOT_EXISTS)

        if isinstance(result, Response):
            return result

        try:
            return jsonify(result)
        except TypeError as e:
            if isinstance(result, Iterable):
                return jsonify(list(result))
            else:
                raise e

    return wrapper


from veripress.api import handlers


def rule(rules, strict_slashes=False, api_func=None, *args, **kwargs):
    """
    Add a API route to the 'api' blueprint.

    :param rules: rule string or string list
    :param strict_slashes: same to Blueprint.route, but default value is False
    :param api_func: a function that returns a JSON serializable object or a Flask Response, or raises ApiException
    :param args: other args that should be passed to Blueprint.route
    :param kwargs: other kwargs that should be passed to Blueprint.route
    :return:
    """
    return url_rule(api_blueprint, rules, strict_slashes=strict_slashes,
                    view_func=json_api(api_func) if api_func else None, *args, **kwargs)


rule('/site', endpoint='site', api_func=handlers.site_info, methods=['GET'])
rule(['/posts',
      '/posts/<int:year>',
      '/posts/<int:year>/<int:month>',
      '/posts/<int:year>/<int:month>/<int:day>',
      '/posts/<int:year>/<int:month>/<int:day>/<string:post_name>'], api_func=handlers.posts, methods=['GET'])
rule('/tags', api_func=handlers.tags, methods=['GET'])
rule('/categories', api_func=handlers.categories, methods=['GET'])
rule('/widgets', api_func=handlers.widgets, methods=['GET'])
rule('/pages/<path:page_path>', api_func=handlers.pages, methods=['GET'], strict_slashes=True)
rule('/search', api_func=handlers.search, methods=['GET'])

rule('/<path:_>', api_func=lambda _: abort(404), methods=['GET'])  # direct all unknown paths to 404



STORAGE_TYPE = '{storage_mode}'
THEME = 'default'
CACHE_TYPE = 'simple'
MODE = 'view-only'  # mixed|api-only|view-only
ENTRIES_PER_PAGE = 5
FEED_COUNT = 10
SHOW_TOC = True
TOC_DEPTH = 3  # 1~6
TOC_LOWEST_LEVEL = 3  # 1~6
ALLOW_SEARCH_PAGES = True
PAGE_SOURCE_ACCESSIBLE = False



import re
from functools import partial

import markdown

from veripress.helpers import to_list


class Parser(object):
    """Base parser class."""

    # this should be overridden in subclasses, and should be a compiled regular expression
    _read_more_exp = None

    def __init__(self):
        if self._read_more_exp is not None and isinstance(self._read_more_exp, str):
            # compile the regular expression
            # make the regex require new lines above and below the sep flag
            self._read_more_exp = re.compile(r'\r?\n\s*?' + self._read_more_exp + r'\s*?\r?\n', re.IGNORECASE)

    def parse_preview(self, raw_content):
        """
        Parse the preview part of the content,
        and return the parsed string and whether there is more content or not.

        If the preview part is equal to the whole part,
        the second element of the returned tuple will be False, else True.

        :param raw_content: raw content
        :return: tuple(parsed string, whether there is more content or not)
        """
        if self._read_more_exp is None:
            return self.parse_whole(raw_content), False

        sp = self._read_more_exp.split(raw_content, maxsplit=1)
        if len(sp) == 2 and sp[0]:
            has_more_content = True
            result = sp[0].rstrip()
        else:
            has_more_content = False
            result = raw_content
        # since the preview part contains no read_more_sep, we can safely use the parse_whole method
        return self.parse_whole(result), has_more_content

    def parse_whole(self, raw_content):
        """Parse the whole part of the content. Should be overridden in subclasses."""
        raise NotImplementedError

    def remove_read_more_sep(self, raw_content):
        """
        Removes the first read_more_sep that occurs in raw_content.
        Subclasses should call this method to preprocess raw_content.
        """
        if self._read_more_exp is None:
            return raw_content

        sp = self._read_more_exp.split(raw_content, maxsplit=1)
        if len(sp) == 2 and sp[0]:
            result = '\n\n'.join((sp[0].rstrip(), sp[1].lstrip()))
        else:
            result = raw_content
        return result


_ext_format_mapping = {}  # key: extension name, value: standard format name
_format_parser_mapping = {}  # key: standard format name, value: parser instance


def get_standard_format_name(ext_name):
    """
    Get the standard format name of the given extension.

    :param ext_name: extension name
    :return: standard format name
    """
    return _ext_format_mapping.get(ext_name.lower())


def get_parser(format_name):
    """
    Get parser of the given format.

    :param format_name: standard format name
    :return: the parser instance
    """
    return _format_parser_mapping.get(format_name.lower())


def parser(format_name, ext_names=None):
    """
    Decorate a parser class to register it.

    :param format_name: standard format name
    :param ext_names: supported extension name
    """

    def decorator(cls):
        format_name_lower = format_name.lower()
        if ext_names is None:
            _ext_format_mapping[format_name_lower] = format_name_lower
        else:
            for ext in to_list(ext_names):
                _ext_format_mapping[ext.lower()] = format_name_lower
        _format_parser_mapping[format_name_lower] = cls()
        return cls

    return decorator


@parser('txt', ext_names=['txt'])
class TxtParser(Parser):
    """Txt content parser."""

    _read_more_exp = r'-{3,}[ \t]*more[ \t]*-{3,}'

    def parse_whole(self, raw_content):
        raw_content = self.remove_read_more_sep(raw_content)
        return '<pre class="txt">{}</pre>'.format(raw_content)


@parser('markdown', ext_names=['md', 'mdown', 'markdown'])
class MarkdownParser(Parser):
    """Markdown content parser."""

    _read_more_exp = r'<!--\s*more\s*-->'

    _markdown = partial(
        markdown.markdown,
        output_format='html5',
        extensions=['markdown.extensions.extra', 'markdown.extensions.codehilite'],
        extension_configs={
            'markdown.extensions.codehilite': {
                'guess_lang': False,
                'css_class': 'highlight',
                'use_pygments': True
            }
        },
    )

    def parse_whole(self, raw_content):
        raw_content = self.remove_read_more_sep(raw_content)
        return self._markdown(raw_content).strip()



from setuptools import setup, find_packages

with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='veripress',
    version='1.0.7',
    packages=find_packages(),
    url='https://github.com/veripress/veripress',
    license='MIT License',
    author='Richard Chien',
    author_email='richardchienthebest@gmail.com',
    description='A blog engine for hackers.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        'Flask', 'Flask-Caching', 'PyYAML', 'markdown',
        'Pygments', 'pytz'
    ],
    python_requires='>=3.4',
    include_package_data=True,
    platforms='any',
    entry_points=dict(
        console_scripts=['veripress=veripress_cli:main']
    )
)



from itertools import islice

from flask import url_for, request, redirect, current_app, send_file, abort, make_response
from werkzeug.contrib.atom import AtomFeed

from veripress import site, cache
from veripress.view import templated, custom_render_template
from veripress.model import storage
from veripress.model.models import Base
from veripress.model.parsers import get_parser
from veripress.helpers import timezone_from_str, parse_toc, validate_custom_page_path


def make_abs_url(unique_key):
    return request.script_root + unique_key


@cache.memoize(timeout=2 * 60)
@templated()
def index(page_num=1):
    if page_num <= 1 and request.path != '/':
        return redirect(url_for('.index'))  # redirect '/page/1' to '/'

    index_page = storage.get_page('index.html', include_draft=False)
    if index_page is not None:
        # there is an 'index.*' custom page, we use this as index.
        return page('index.html')

    all_posts = storage.get_posts(include_draft=False)

    count = current_app.config['ENTRIES_PER_PAGE']
    start = (page_num - 1) * count

    posts = []
    for post_ in islice(all_posts, start, start + count + 1):  # slice an additional one to check if there is more
        post_d = post_.to_dict()
        del post_d['raw_content']
        post_d['preview'], post_d['has_more_content'] = get_parser(post_.format).parse_preview(post_.raw_content)
        post_d['url'] = make_abs_url(post_.unique_key)
        posts.append(post_d)

    if start > 0:
        next_page_num = page_num - 1
        next_url = url_for('.index', page_num=next_page_num if next_page_num != 1 else None)
    else:
        next_url = None
    if len(posts) > count:
        # the additional one is included
        posts = posts[:count]
        prev_url = url_for('.index', page_num=page_num + 1)
    else:
        prev_url = None

    return dict(entries=posts, next_url=next_url, prev_url=prev_url)


@cache.memoize(timeout=2 * 60)
@templated()
def post(year, month, day, post_name):
    rel_url = request.path[len('/post/'):]
    fixed_rel_url = storage.fix_post_relative_url(rel_url)
    if rel_url != fixed_rel_url:
        return redirect(request.url_root + 'post/' + fixed_rel_url)  # it's not the correct relative url, so redirect

    post_ = storage.get_post(rel_url, include_draft=False)
    if post_ is None:
        abort(404)

    post_d = post_.to_dict()
    del post_d['raw_content']
    post_d['content'] = get_parser(post_.format).parse_whole(post_.raw_content)
    post_d['content'], post_d['toc'], post_d['toc_html'] = parse_toc(post_d['content'])
    post_d['url'] = make_abs_url(post_.unique_key)
    post_ = post_d

    return custom_render_template(post_['layout'] + '.html', entry=post_)


@templated()
def page(rel_url):
    if not validate_custom_page_path(rel_url):
        abort(403)

    fixed_rel_url, exists = storage.fix_page_relative_url(rel_url)
    if exists:
        file_path = fixed_rel_url
        return send_file(file_path)  # send direct file
    elif fixed_rel_url is None:  # relative url is invalid
        abort(404)  # pragma: no cover, this is never possible when visiting this site in web browser
    elif rel_url != fixed_rel_url:
        return redirect(url_for('.page', rel_url=fixed_rel_url))  # it's not the correct relative url, so redirect

    resp = cache.get('view-handler.' + rel_url)
    if resp is not None:
        return resp  # pragma: no cover, here just get the cached response

    page_ = storage.get_page(rel_url, include_draft=False)
    if page_ is None:
        abort(404)

    page_d = page_.to_dict()
    del page_d['raw_content']
    page_d['content'] = get_parser(page_.format).parse_whole(page_.raw_content)
    page_d['content'], page_d['toc'], page_d['toc_html'] = parse_toc(page_d['content'])
    page_d['url'] = make_abs_url(page_.unique_key)
    page_ = page_d

    resp = custom_render_template(page_['layout'] + '.html', entry=page_)
    cache.set('view-handler.' + rel_url, resp, timeout=2 * 60)
    return resp


@cache.memoize(timeout=2 * 60)
@templated('category.html', 'archive.html')
def category(category_name):
    posts = storage.get_posts_with_limits(include_draft=False, **{'categories': [category_name]})
    if not posts:
        abort(404)

    def convert_to_dict(post_):
        post_d = post_.to_dict()
        del post_d['raw_content']
        post_d['preview'], post_d['has_more_content'] = get_parser(post_.format).parse_preview(post_.raw_content)
        post_d['url'] = make_abs_url(post_.unique_key)
        return post_d

    posts = list(map(convert_to_dict, posts))
    return dict(entries=posts, archive_type='Category', archive_name=category_name)


@cache.memoize(timeout=2 * 60)
@templated('tag.html', 'archive.html')
def tag(tag_name):
    posts = storage.get_posts_with_limits(include_draft=False, **{'tags': [tag_name]})
    if not posts:
        abort(404)

    def convert_to_dict(post_):
        post_d = post_.to_dict()
        del post_d['raw_content']
        post_d['preview'], post_d['has_more_content'] = get_parser(post_.format).parse_preview(post_.raw_content)
        post_d['url'] = make_abs_url(post_.unique_key)
        return post_d

    posts = list(map(convert_to_dict, posts))
    return dict(entries=posts, archive_type='Tag', archive_name=tag_name)


@cache.memoize(timeout=2 * 60)
@templated()
def archive(year=None, month=None):
    posts = storage.get_posts_with_limits(include_draft=False)

    rel_url_prefix = ''
    archive_name = ''
    if year is not None:
        rel_url_prefix += '%04d/' % year
        archive_name += str(year)
    if month is not None:
        rel_url_prefix += '%02d/' % month
        archive_name += '.' + str(month)

    def convert_to_dict(post_):
        post_d = post_.to_dict()
        del post_d['raw_content']
        post_d['preview'], post_d['has_more_content'] = get_parser(post_.format).parse_preview(post_.raw_content)
        post_d['url'] = make_abs_url(post_.unique_key)
        return post_d

    posts = list(map(convert_to_dict, filter(lambda p: p.rel_url.startswith(rel_url_prefix), posts)))
    return dict(entries=posts, archive_type='Archive', archive_name=archive_name if archive_name else 'All')


@templated('search.html', 'archive.html')
def search():
    raw_query = request.args.get('q', '').strip()
    query = raw_query.lower()
    if not query:
        abort(404)

    def process(p):
        del p['raw_content']
        p['url'] = make_abs_url(p['unique_key'])
        return p

    result = list(map(process, map(Base.to_dict, storage.search_for(query))))
    return dict(entries=result, archive_type='Search', archive_name='"{}"'.format(raw_query))


@cache.memoize(timeout=2 * 60)
def feed():
    def convert_to_dict(p):
        post_d = p.to_dict()
        del post_d['raw_content']
        post_d['content'] = get_parser(p.format).parse_whole(p.raw_content)
        post_d['url'] = site['root_url'] + make_abs_url(p.unique_key)
        return post_d

    posts = map(convert_to_dict, islice(storage.get_posts(include_draft=False), 0, current_app.config['FEED_COUNT']))

    atom = AtomFeed(title=site['title'],
                    subtitle=site['subtitle'],
                    url=site['root_url'] + request.script_root,
                    feed_url=site['root_url'] + url_for('.feed'),
                    author=site.get('author'))
    for post_ in posts:
        atom.add(title=post_['title'],
                 content=post_['content'],
                 url=post_['url'],
                 id=post_['unique_key'],
                 published=post_['created'].replace(tzinfo=timezone_from_str(site['timezone'])),
                 updated=post_['updated'].replace(tzinfo=timezone_from_str(site['timezone'])),
                 author=post_['author'])

    response = make_response(atom.to_string())
    response.content_type = 'application/atom+xml; charset=utf-8'
    return response



import click


@click.group(name='veripress',
             short_help='A blog engine for hackers.',
             help='This is a blog engine for hackers. '
                  'You can use this to serve a blog, a wiki or anything else you like.')
@click.version_option(version='1.0.6')
def cli():
    pass


def main():
    cli.main()


import veripress_cli.init
import veripress_cli.serve
import veripress_cli.theme
import veripress_cli.generate
import veripress_cli.deploy



import os
import shutil
from datetime import datetime

import click

from veripress_cli import cli


@cli.command('init', short_help='Initialize a new instance directory.',
             help='This command will initialize the current working directory as a new VeriPress instance, '
                  'which means to create default configuration file, necessary subdirectories, etc.')
@click.option('--storage-mode', '-s', default='file', type=click.Choice(['file']),
              help='Storage mode (only "file" mode supported currently).')
def init_command(storage_mode):
    from veripress import app
    instance_path = app.instance_path
    defaults_dir = os.path.join(os.path.dirname(__file__), 'defaults')

    with open(os.path.join(defaults_dir, 'config.py'), 'r', encoding='utf-8') as f_default_conf:
        with open(os.path.join(instance_path, 'config.py'), 'w', encoding='utf-8') as f_conf:
            f_conf.write(f_default_conf.read().format(storage_mode=storage_mode))

    shutil.copyfile(os.path.join(defaults_dir, 'site.json'),
                    os.path.join(instance_path, 'site.json'))
    shutil.copytree(os.path.join(defaults_dir, 'static'),
                    os.path.join(instance_path, 'static'))
    os.mkdir(os.path.join(instance_path, 'themes'))

    if storage_mode == 'file':
        init_file_storage(instance_path)

    click.echo('\nDefault files and configurations has been created!\n\n'
               'Now you can run "veripress theme install default" to install the default theme '
               'and then run "veripress preview" to preview the blog.\n\nEnjoy!')


def init_file_storage(instance_path):
    os.mkdir(os.path.join(instance_path, 'posts'))
    os.mkdir(os.path.join(instance_path, 'pages'))
    os.mkdir(os.path.join(instance_path, 'widgets'))

    now_dt = datetime.now()
    first_post_file_path = os.path.join(instance_path, 'posts', '{}-hello-world.md'.format(now_dt.strftime('%Y-%m-%d')))
    with open(first_post_file_path, 'w', encoding='utf-8') as f:
        f.write('---\ntitle: Hello, world!\ncreated: {}\n'
                'categories: Hello\ntags: [Hello, Greeting]\n---\n\n'.format(now_dt.strftime('%Y-%m-%d %H:%M:%S')))
        f.write('This is the hello world post!\n')

    os.mkdir(os.path.join(instance_path, 'pages', 'hello'))
    first_page_file_path = os.path.join(instance_path, 'pages', 'hello', 'index.md')
    with open(first_page_file_path, 'w', encoding='utf-8') as f:
        f.write('---\ntitle: Hello\ncreated: {}\n---\n\n'.format(now_dt.strftime('%Y-%m-%d %H:%M:%S')))
        f.write('This is the hello custom page!\n')

    first_widget_file_path = os.path.join(instance_path, 'widgets', 'welcome.md')
    with open(first_widget_file_path, 'w', encoding='utf-8') as f:
        f.write('---\nposition: sidebar\norder: 0\n---\n\n')
        f.write('#### Welcome!\n\nHi! Welcome to my blog.\n')
